from __future__ import annotations

from .editor import NodeEditorObject, EclipseNodeEditor
from PyQt6.QtGui import QColor, QFont
import math


class SocketConnection:
    def __init__(self):
        self.from_socket = None
        self.to_socket = None


class Socket(NodeEditorObject):
    def __init__(self):
        super().__init__()
        self.connected_height = 25
        self.width = 0
        self.height = 25
        self.pin_radius = 5
        self.pin_hover_radius = 7
        self.pin_color = QColor(255, 255, 255)
        self.pin_type = "input"
        self.connections: list[SocketConnection] = []
        self.is_hovered = False
        self.is_pin_hovered = False

    def render(self, node: Node, x: int, y: int, editor: EclipseNodeEditor):
        pin_x = x
        if self.pin_type == "input":
            pin_x = x

        if self.pin_type == "output":
            pin_x = x + node.width

        editor.draw_circle(pin_x, y + self.connected_height / 2, self.pin_hover_radius if self.is_pin_hovered else self.pin_radius,
                           self.pin_color.red() / 255,
                           self.pin_color.green() / 255,
                           self.pin_color.blue() / 255)

    def check_show_interface(self):
        if self.pin_type == "output":
            return False

        if self.pin_type == "input":
            if self.connections:
                return False

        return True


class Node(NodeEditorObject):
    def __init__(self, x=0, y=0):

        super().__init__()
        # Position and size
        self.x = x
        self.y = y
        self.width = 100
        self.height = 100

        # Colors
        self.primary_color = QColor(75, 75, 75)    # Header color
        self.secondary_color = QColor(50, 50, 50)  # Background color
        self.title_color = QColor(255, 255, 255)   # Title text color

        # Title properties
        self.title = "Node"
        self.title_font_size = 12
        self.header_size = 30

        # State
        self.is_selected = False
        self.is_hovered = False

        # Sockets
        self.sockets: list[Socket] = []
        self.sockets_spacing = 2
        self.render_socket_regions = False

    def add_input_socket(self, socket: Socket):
        socket.pin_type = "input"
        self.sockets.append(socket)
        return socket

    def add_output_socket(self, socket: Socket):
        socket.pin_type = "output"
        self.sockets.append(socket)
        return socket

    def input(self, editor: EclipseNodeEditor):
        """Handle input for the node"""
        mouse_pos = editor.world_mouse_pos
        # Normalize the mouse position into the nodes shape for a UV like coordinate system
        normalized_x = (mouse_pos.x() - self.x) / self.width

        normalized_y = (mouse_pos.y() - self.y) / self.height

        # Adjust the mouse position to the node's position so that top left is (0, 0)
        mouse_x = mouse_pos.x()
        mouse_y = mouse_pos.y()

        # Check if the mouse is over the node
        self.is_hovered = (
            normalized_x >= 0 and normalized_x <= 1 and
            normalized_y >= 0 and normalized_y <= 1
        )

        # Check if the mouse is over any sockets
        socket_y = self.y + self.header_size
        for i, socket in enumerate(self.sockets):
            socket.is_pin_hovered = False
            socket.is_hovered = False
            if mouse_x >= self.x and mouse_x <= self.x + self.width and mouse_y >= socket_y and mouse_y <= socket_y + socket.height:
                socket.is_hovered = True

            if math.sqrt((mouse_x - self.get_socket_x(socket)) ** 2 + (mouse_y - (socket_y + socket.connected_height / 2)) ** 2) <= socket.pin_hover_radius:
                socket.is_pin_hovered = True


            show_interface = socket.check_show_interface()
            if show_interface:

                socket_y += socket.height + self.sockets_spacing
            else:
                socket_y += socket.connected_height + self.sockets_spacing

    def update(self, editor: EclipseNodeEditor):
        """Update node state"""
        width_needed = self.width
        height_needed = self.header_size
        for socket in self.sockets:

            width_needed = max(width_needed, socket.width)
            show_interface = socket.check_show_interface()
            if show_interface:
                height_needed += socket.height + self.sockets_spacing
            else:
                height_needed += socket.connected_height + self.sockets_spacing

        self.width = width_needed
        self.height = height_needed

    def render(self, editor):
        """Render the node"""
        # Draw outer shadow first (larger and more transparent)
        outer_shadow_offset = 6
        outer_shadow_expand = 4
        editor.draw_rectangle(
            self.x - outer_shadow_expand,
            self.y + outer_shadow_offset - outer_shadow_expand - 6,
            self.width + outer_shadow_expand * 2,
            self.height + outer_shadow_expand * 4,
            0, 0, 0, 0.15  # Very transparent black
        )

        # Draw inner shadow (smaller and more opaque)
        inner_shadow_offset = 4
        inner_shadow_expand = 2
        editor.draw_rectangle(
            self.x - inner_shadow_expand,
            self.y + inner_shadow_offset - inner_shadow_expand - 4,
            self.width + inner_shadow_expand * 2,
            self.height + inner_shadow_expand * 4,
            0, 0, 0, 0.4  # Semi-transparent black

        )

        # Draw node background (main body)
        editor.draw_rectangle(
            self.x,
            self.y,
            self.width,
            self.height,
            self.secondary_color.red() / 255,
            self.secondary_color.green() / 255,
            self.secondary_color.blue() / 255
        )

        # Draw node header
        editor.draw_rectangle(
            self.x,
            self.y,
            self.width,
            self.header_size,
            self.primary_color.red() / 255,
            self.primary_color.green() / 255,
            self.primary_color.blue() / 255
        )

        # Draw title text
        # Center the text vertically in the header
        text_x = self.x + 10  # Add some padding from the left
        text_y = self.y + (self.header_size / 2)  # Center point of header

        editor.draw_text(
            text_x,
            text_y,
            self.title,
            self.title_color.red() / 255,
            self.title_color.green() / 255,
            self.title_color.blue() / 255,
            1.0,
            self.title_font_size
        )

        # Render sockets
        socket_y = self.y + self.header_size
        for i, socket in enumerate(self.sockets):
            if socket.is_hovered:
                editor.draw_rectangle(
                    self.x,
                    socket_y,
                    self.width,
                    socket.height,
                    1, 1, 1, 0.03
                )
            if self.render_socket_regions:
                clr_r = 1
                clr_g = 0
                clr_b = 0
                if i % 2 == 1:
                    clr_r = 0
                    clr_g = 1
                    clr_b = 1
                editor.draw_rectangle(
                    self.x,
                    socket_y,
                    self.width,
                    socket.height,
                    clr_r, clr_g, clr_b, 0.03
                )

            socket.render(self, self.x, socket_y, editor)
            show_interface = socket.check_show_interface()
            if show_interface:
                socket_y += socket.height + self.sockets_spacing

            else:
                socket_y += socket.connected_height + self.sockets_spacing

    def get_socket_x(self, socket: Socket):
        if socket.pin_type == "input":
            return self.x
        if socket.pin_type == "output":
            return self.x + self.width
