from PyQt6.QtOpenGLWidgets import QOpenGLWidget
from PyQt6.QtCore import Qt, QSize, QPoint, QTimer
from PyQt6.QtGui import QColor, QFont, QPainter
from OpenGL.GL import *
from abc import ABC, abstractmethod
import math


class NodeEditorObject():
    def __init__(self):
        pass

    def input(self, editor):
        """Handle input for this object"""
        pass

    def update(self, editor):
        """Update object state"""
        pass

    def render(self, editor):
        """Render the object"""
        pass


class EclipseNodeEditor(QOpenGLWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.background_color = QColor(21, 21, 21)
        # Set update behavior to redraw continuously
        self.setUpdateBehavior(QOpenGLWidget.UpdateBehavior.NoPartialUpdate)

        # Enable mouse tracking for smooth pan updates
        self.setMouseTracking(True)

        # Initialize view transform (current values)
        self.pan_x = 0
        self.pan_y = 0
        self.zoom = 1.0

        # Target transform values
        self.target_pan_x = 0
        self.target_pan_y = 0
        self.target_zoom = 1.0

        # Transform constraints
        self.min_zoom = 0.1
        self.max_zoom = 10.0

        # Interpolation settings
        self.lerp_factor = 0.2

        # Mouse state
        self.mouse_pos = QPoint()
        self.world_mouse_pos = QPoint()
        self.last_mouse_drag_pos = QPoint()
        self.is_panning = False

        # Objects list
        self.objects = []

        # QPainter for text rendering
        self.painter = QPainter()

        # Setup update timer for smooth interpolation and object updates
        self.update_timer = QTimer(self)
        self.update_timer.timeout.connect(self.on_timer_update)
        self.update_timer.start(16)  # ~60 FPS

    def add_object(self, obj):
        """Add an object to the editor"""
        self.objects.append(obj)

    def remove_object(self, obj):
        """Remove an object from the editor"""
        if obj in self.objects:
            self.objects.remove(obj)

    def on_timer_update(self):
        """Handle timer updates for transform interpolation and object updates"""
        # Update current mouse position
        self.mouse_pos = self.mapFromGlobal(self.cursor().pos())

        # Update world mouse position
        world_x, world_y = self.screen_to_world(
            self.mouse_pos.x(), self.mouse_pos.y())
        self.world_mouse_pos.setX(int(world_x))
        self.world_mouse_pos.setY(int(world_y))

        # First handle transform interpolation
        self.interpolate_transform()

        # Then update all objects
        self.process_objects()

    def process_objects(self):
        """Process input and updates for all objects"""
        # First process all inputs
        for obj in self.objects:
            obj.input(self)

        # Then process all updates
        for obj in self.objects:
            obj.update(self)

    def render_objects(self):
        """Render all objects"""
        for obj in self.objects:
            obj.render(self)

    def lerp(self, start, end, factor):
        """Linear interpolation between start and end values"""
        return start + (end - start) * factor

    def interpolate_transform(self):
        """Interpolate between current and target transform values"""
        # Check if we need to update pan
        if abs(self.target_pan_x - self.pan_x) > 0.01 or abs(self.target_pan_y - self.pan_y) > 0.01:
            self.pan_x = self.lerp(
                self.pan_x, self.target_pan_x, self.lerp_factor)
            self.pan_y = self.lerp(
                self.pan_y, self.target_pan_y, self.lerp_factor)

        # Check if we need to update zoom
        if abs(self.target_zoom - self.zoom) > 0.0001:
            self.zoom = self.lerp(
                self.zoom, self.target_zoom, self.lerp_factor)

        # Always update to ensure continuous frame updates
        self.update()

    def initializeGL(self):
        """Initialize OpenGL settings"""
        glClearColor(21/255, 21/255, 21/255, 1.0)
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)

    def draw_rectangle(self, x, y, width, height, r, g, b, a=1.0):
        """Helper method to draw a rectangle"""
        glColor4f(r, g, b, a)
        glBegin(GL_QUADS)
        glVertex2f(x, y)
        glVertex2f(x + width, y)
        glVertex2f(x + width, y + height)
        glVertex2f(x, y + height)
        glEnd()

    def draw_text(self, x, y, text, r, g, b, a=1.0, font_size=12):
        """Helper method to draw text using Qt's painter"""
        # Start QPainter
        self.painter.begin(self)
        self.painter.setRenderHint(QPainter.RenderHint.Antialiasing)

        # Set up the font
        font = QFont()
        font.setPointSizeF(font_size * self.zoom)
        self.painter.setFont(font)

        # Set color
        self.painter.setPen(
            QColor(int(r * 255), int(g * 255), int(b * 255), int(a * 255)))

        # Convert position to screen space
        screen_x = x * self.zoom + self.pan_x
        screen_y = y * self.zoom + self.pan_y

        # Get font metrics for proper vertical centering
        metrics = self.painter.fontMetrics()
        text_height = metrics.height()

        # Draw the text centered vertically (y is the center point)
        # Subtract half the text height to center it vertically
        self.painter.drawText(int(screen_x), int(
            screen_y - text_height/2 + metrics.ascent()), text)

        # End QPainter
        self.painter.end()

        # Restore OpenGL state
        glEnable(GL_BLEND)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
        self.apply_transform()

    def draw_circle(self, x, y, radius, r, g, b, a=1.0):
        """Helper method to draw a circle"""
        glColor4f(r, g, b, a)
        glBegin(GL_TRIANGLE_FAN)
        glVertex2f(x, y)  # Center point
        segments = 32  # Number of segments to make the circle smooth
        for i in range(segments + 1):
            angle = 2.0 * math.pi * i / segments
            dx = radius * math.cos(angle)
            dy = radius * math.sin(angle)
            glVertex2f(x + dx, y + dy)
        glEnd()

    def screen_to_world(self, screen_x, screen_y):
        """Convert screen coordinates to world coordinates"""
        world_x = (screen_x - self.pan_x) / self.zoom
        world_y = (screen_y - self.pan_y) / self.zoom
        return world_x, world_y

    def world_to_screen(self, world_x, world_y):
        """Convert world coordinates to screen coordinates"""
        screen_x = world_x * self.zoom + self.pan_x
        screen_y = world_y * self.zoom + self.pan_y
        return screen_x, screen_y

    def apply_transform(self):
        """Apply the current view transform"""
        width = self.width()
        height = self.height()

        # Set up orthographic projection
        glLoadIdentity()
        glOrtho(0, width, height, 0, -1, 1)

        # Apply transforms in correct order
        glTranslatef(self.pan_x, self.pan_y, 0)
        glScalef(self.zoom, self.zoom, 1.0)

    def paintGL(self):
        """Paint the scene"""
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # Apply the view transform
        self.apply_transform()

        # Render all objects
        self.render_objects()

    def resizeGL(self, width, height):
        """Handle window resize"""
        glViewport(0, 0, width, height)

    def mousePressEvent(self, event):
        """Handle mouse press events"""
        if event.button() == Qt.MouseButton.MiddleButton:
            self.is_panning = True
            self.last_mouse_drag_pos = event.pos()
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """Handle mouse release events"""
        if event.button() == Qt.MouseButton.MiddleButton:
            self.is_panning = False
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):
        """Handle mouse move events"""
        # Update current mouse position
        if self.is_panning:
            delta = event.pos() - self.last_mouse_drag_pos
            self.target_pan_x += delta.x()
            self.target_pan_y += delta.y()
            self.last_mouse_drag_pos = event.pos()
        super().mouseMoveEvent(event)

    def wheelEvent(self, event):
        """Handle mouse wheel events for zooming"""
        # Get mouse position
        mouse_pos = event.position()
        mouse_x = mouse_pos.x()
        mouse_y = mouse_pos.y()

        # Get the world point under mouse before zoom
        old_world_x, old_world_y = self.screen_to_world(mouse_x, mouse_y)

        # Calculate and clamp new zoom level
        zoom_factor = 1.2 if event.angleDelta().y() > 0 else 1/1.2
        new_zoom = self.target_zoom * zoom_factor

        if self.min_zoom <= new_zoom <= self.max_zoom:
            # Set target zoom
            self.target_zoom = new_zoom

            # Calculate the screen point for this world position at target zoom
            # We need to use current pan but target zoom for this calculation
            screen_x = old_world_x * new_zoom + self.target_pan_x
            screen_y = old_world_y * new_zoom + self.target_pan_y

            # Set target pan to keep world point under mouse
            self.target_pan_x += mouse_x - screen_x
            self.target_pan_y += mouse_y - screen_y

        event.accept()

    def minimumSizeHint(self):
        """Ensure the widget has a minimum size"""
        return QSize(50, 50)

    def sizeHint(self):
        """Preferred size for the widget"""
        return QSize(400, 400)
